import logging
import warnings

import numpy as np

import pandas as pd


from itertools import combinations
from ..utils.connectivity import face_face_connectivity

logger = logging.getLogger(name=__name__)


def split_vert(sheet, vert, face, to_rewire, epsilon, recenter=False):
    """ Creates a new vertex and moves it towards the center of face.

    The edges in to_rewire will be connected to the new vertex.

    Parameters
    ----------

    sheet : a :class:`tyssue.Sheet` instance
    vert : int, the index of the vertex to split
    face : int, the index of the face where to move the vertex
    to_rewire : :class:`pd.DataFrame` a subset of `sheet.edge_df`
        where all the edges pointing to (or from) the old vertex will point
        to (or from) the new.

    Note
    ----

    This will leave opened faces and cells

    """
    # Add a vertex
    sheet.vert_df = sheet.vert_df.append(sheet.vert_df.loc[vert], ignore_index=True)
    new_vert = sheet.vert_df.index[-1]

    # Move it towards the face center
    r_ia = sheet.face_df.loc[face, sheet.coords] - sheet.vert_df.loc[vert, sheet.coords]
    if recenter:
        sheet.vert_df.loc[new_vert, sheet.coords] += (
            0.5 * r_ia * epsilon / np.linalg.norm(r_ia)
        )
        sheet.vert_df.loc[vert, sheet.coords] -= (
            0.5 * r_ia * epsilon / np.linalg.norm(r_ia)
        )
    else:
        sheet.vert_df.loc[new_vert, sheet.coords] += (
            r_ia * epsilon / np.linalg.norm(r_ia)
        )

    # rewire
    sheet.edge_df.loc[to_rewire.index] = to_rewire.replace(
        {"srce": vert, "trgt": vert}, new_vert
    )


def add_vert(eptm, edge):
    """Adds a vertex in the middle of the edge,

    which is split as is its opposite(s)

    Parameters
    ----------
    eptm : a :class:`Epithelium` instance
    edge : int
    the index of one of the half-edges to split

    Returns
    -------
    new_vert : int
    the index to the new vertex
    new_edges : int or list of ints
    index to the new edge(s). For a sheet, returns
    a single index, for a 3D epithelium, returns
    the list of all the new parallel edges
    new_opp_edges : int or list of ints
    index to the new opposite edge(s). For a sheet, returns
    a single index, for a 3D epithelium, returns
    the list of all the new parallel edges


    In the simple case whith two half-edge, returns
    indices to the new edges, with the following convention:

    s    e    t
      ------>
    * <------ *
    oe

    s    e       ne   t
      ------   ----->
    * <----- * ------ *
        oe   nv   noe

    where "e" is the passed edge as argument, "s" its source "t" its
    target and "oe" its opposite. The returned edges are the ones
    between the new vertex and the input edge's original target.
    """

    srce, trgt = eptm.edge_df.loc[edge, ["srce", "trgt"]]
    opposites = eptm.edge_df[
        (eptm.edge_df["srce"] == trgt) & (eptm.edge_df["trgt"] == srce)
    ]
    parallels = eptm.edge_df[
        (eptm.edge_df["srce"] == srce) & (eptm.edge_df["trgt"] == trgt)
    ]

    new_vert = eptm.vert_df.loc[[srce, trgt]].mean()
    eptm.vert_df = eptm.vert_df.append(new_vert, ignore_index=True)
    new_vert = eptm.vert_df.index[-1]
    new_edges = []

    for p, p_data in parallels.iterrows():
        eptm.edge_df.loc[p, "trgt"] = new_vert
        eptm.edge_df = eptm.edge_df.append(p_data, ignore_index=True)
        new_edge = eptm.edge_df.index[-1]
        eptm.edge_df.loc[new_edge, "srce"] = new_vert
        eptm.edge_df.loc[new_edge, "trgt"] = trgt
        new_edges.append(new_edge)

    new_opp_edges = []
    for o, o_data in opposites.iterrows():
        eptm.edge_df.loc[o, "srce"] = new_vert
        eptm.edge_df = eptm.edge_df.append(o_data, ignore_index=True)
        new_opp_edge = eptm.edge_df.index[-1]
        eptm.edge_df.loc[new_opp_edge, "trgt"] = new_vert
        eptm.edge_df.loc[new_opp_edge, "srce"] = trgt
        new_opp_edges.append(new_opp_edge)

    # ## Sheet special case
    if len(new_edges) == 1:
        new_edges = new_edges[0]
    if len(new_opp_edges) == 1:
        new_opp_edges = new_opp_edges[0]
    elif len(new_opp_edges) == 0:
        new_opp_edges = None
    return new_vert, new_edges, new_opp_edges


def close_face(eptm, face):
    """Closes the face if a single edge is missing.

    This function **does not** close the adjacent and opposite
    faces.
    """
    face_edges = eptm.edge_df[eptm.edge_df["face"] == face]
    srces = set(face_edges["srce"])
    trgts = set(face_edges["trgt"])

    if srces == trgts:
        logger.debug("Face {} already closed".format(face))
        return
    try:
        single_srce, = srces.difference(trgts)
        single_trgt, = trgts.difference(srces)
    except ValueError as err:
        print("Closing only possible with exactly two dangling vertices")
        raise err

    eptm.edge_df = eptm.edge_df.append(face_edges.iloc[0], ignore_index=True)
    eptm.edge_df.index.name = "edge"
    new_edge = eptm.edge_df.index[-1]
    eptm.edge_df.loc[new_edge, ["srce", "trgt"]] = single_trgt, single_srce


def drop_two_sided_faces(eptm):
    """Removes all the two (or one?) sided faces from the epithelium

    Note that they are not collapsed, but simply eliminated
    Does not reindex
    """

    num_sides = eptm.edge_df.groupby("face").size()
    if num_sides.min() > 2:
        return

    two_sided = eptm.face_df[num_sides < 3].index
    edges = eptm.edge_df[eptm.edge_df["face"].isin(two_sided)].index
    eptm.edge_df.drop(edges, axis=0, inplace=True)
    eptm.face_df.drop(two_sided, axis=0, inplace=True)


def remove_face(sheet, face):
    """Removes a face from the mesh
    """

    edges = sheet.edge_df[sheet.edge_df["face"] == face]
    verts = edges["srce"].unique()
    new_vert_data = sheet.vert_df.loc[verts].mean()
    sheet.vert_df = sheet.vert_df.append(new_vert_data, ignore_index=True)
    new_vert = sheet.vert_df.index[-1]

    # collapse all edges connected to the face vertices
    sheet.edge_df.replace({"srce": verts, "trgt": verts}, new_vert, inplace=True)

    collapsed = sheet.edge_df.query("srce == trgt")

    sheet.edge_df.drop(collapsed.index, axis=0, inplace=True)
    remanent = sheet.edge_df.query(f"face == {face}").index
    if remanent.shape[0]:
        warnings.warn(f"something fishy with face {face}")
        sheet.edge_df.drop(remanent, axis=0, inplace=True)

    sheet.face_df.drop(face, axis=0, inplace=True)
    sheet.vert_df.drop(verts, axis=0, inplace=True)

    logger.info("removed %d of %d vertices", len(verts), sheet.vert_df.shape[0])
    logger.info("face %d is now dead ", face)
    drop_two_sided_faces(sheet)

    sheet.reset_index()
    sheet.reset_topo()

    return 0


def collapse_edge(sheet, edge, reindex=True, allow_two_sided=False):
    """Collapses edge and merges it's vertices, creating (or increasing the rank of)
    a rosette structure.

    If `reindex` is `True` (the default), resets indexes and topology data.
    The edge is collapsed on the smaller of the srce, trgt indexes (to minimize reindexing impact)

    """
    srce, trgt = np.sort(sheet.edge_df.loc[edge, ["srce", "trgt"]]).astype(int)

    edges = sheet.edge_df[
        ((sheet.edge_df["srce"] == srce) & (sheet.edge_df["trgt"] == trgt))
        | ((sheet.edge_df["srce"] == trgt) & (sheet.edge_df["trgt"] == srce))
    ]

    has_3_sides = np.any(sheet.face_df.loc[edges["face"].astype(int), "num_sides"] < 4)
    if has_3_sides and not allow_two_sided:
        warnings.warn("""This operation would result in a two sided face, aborting""")
        return -1

    sheet.vert_df.loc[srce, sheet.coords] = sheet.vert_df.loc[
        [srce, trgt], sheet.coords
    ].mean(axis=0)
    sheet.vert_df.drop(trgt, axis=0, inplace=True)
    # rewire
    sheet.edge_df.replace({"srce": trgt, "trgt": trgt}, srce, inplace=True)
    # all the edges parallel to the original
    collapsed = sheet.edge_df.query("srce == trgt")
    sheet.edge_df.drop(collapsed.index, axis=0, inplace=True)
    if allow_two_sided:
        drop_two_sided_faces(sheet)
    if reindex:
        sheet.reset_index()
        sheet.reset_topo()
    return 0


def merge_vertices(sheet, vert0, vert1, reindex=True):
    """Merge the two vertices vert0 and vert1 iff they are linked by an edge

    If `reindex` is `True` (the default), resets indexes and topology data

    Note:
    -----
    It is more efficient to call directly `collapse_edge`

    """
    edges = sheet.edge_df[
        ((sheet.edge_df["srce"] == vert0) & (sheet.edge_df["trgt"] == vert1))
        | ((sheet.edge_df["srce"] == vert1) & (sheet.edge_df["trgt"] == vert0))
    ].index
    if not len(edges):
        raise ValueError(
            f"""No edge found between vertices {vert0} and {vert1}, cannot merge"""
        )
    return collapse_edge(sheet, edges[0], reindex)


def condition_4i(eptm):
    """
    Return an index over the faces violating condition 4 i in Okuda et al 2013,
    that is edges (from the same face) sharing two vertices simultaneously.
    """
    num_srces = eptm.edge_df.groupby("face")["srce"].apply(lambda s: len(set(s)))
    num_sides = eptm.face_df["num_sides"]
    return eptm.face_df[(num_srces != num_sides) | (num_sides < 3)].index


def get_neighbour_face_pairs(eptm):
    """
    Returns a pandas Series of neighboring face pairs (as forzen sets of 2 indexes)
    """
    pairs = []
    eptm.edge_df["v_pair"] = eptm.edge_df[["srce", "trgt"]].apply(frozenset, axis=1)

    _ = eptm.edge_df.groupby("v_pair")["face"].apply(
        lambda s: pairs.extend(
            [frozenset((a, b)) for a, b in combinations(s.values, 2)]
        )
    )
    return pd.Series(pairs).drop_duplicates()


def get_num_common_edges(eptm):
    """
    Returns the number of common edges between two neighboring faces
    this number is set to -1 if those faces are opposite and share the
    same edges.
    """
    pairs = get_neighbour_face_pairs(eptm)
    face_v_pair_orbit = eptm.edge_df.groupby("face").apply(
        lambda df: frozenset(df["v_pair"])
    )
    n_common = [
        len(face_v_pair_orbit.loc[fa].intersection(face_v_pair_orbit.loc[fb]))
        if face_v_pair_orbit.loc[fb] != face_v_pair_orbit.loc[fa]
        else -1
        for fa, fb in pairs
    ]
    n_common = pd.Series(n_common, index=pd.Index(pairs, name="face_pairs"))
    return n_common


def condition_4ii(eptm):
    """
    Return an array of face pairs sharing more than two half-edges, as defined
    in Okuda et al. 2013 condition 4 ii
    """
    conmat = face_face_connectivity(eptm, exclude_opposites=True)
    return np.vstack(np.where(conmat > 2)).T


def merge_border_edges(sheet, drop_two_sided=True):
    """Merge edges at the border of a sheet such that no vertex has only
    one incoming and one outgoing edge.

    """

    single_trgt = sheet.edge_df[
        sheet.upcast_trgt(sheet.edge_df.groupby("trgt").apply(len) == 1)
    ]
    faces = set(single_trgt["face"])
    single_srce = sheet.edge_df[
        sheet.upcast_srce(sheet.edge_df.groupby("srce").apply(len) == 1)
    ]
    sheet.edge_df.drop(single_srce.index, inplace=True)
    sheet.edge_df.drop(
        set(single_trgt.index).difference(single_srce.index), inplace=True
    )
    for face in faces:
        close_face(sheet, face)

    if drop_two_sided:
        drop_two_sided_faces(sheet)

    sheet.reset_index()
    sheet.reset_topo()



"""
Small module to manage units and dimension analysis
"""
import quantities as pq


temperature = 300 * pq.K
kT = pq.UnitQuantity("k_B T", definition=300 * pq.K * 1 * pq.constants.k, symbol="kT")

length = pq.um
force = pq.UnitQuantity("nanonewton", definition=1e-9 * pq.N, symbol="nN")
energy = pq.UnitQuantity("femtojoules", definition=force * length, symbol="fJ")
time = pq.s

line_tension = energy / length
line_elasticity = energy / length ** 2
line_viscosity = force * time / length


area = length ** 2
area_tension = energy / area
area_elasticity = energy / area ** 2

vol = length ** 3
vol_tension = energy / vol
vol_elasticity = energy / vol ** 2



"""Connectivity matrix computation
"""

import pandas as pd
import numpy as np

from scipy import sparse


def _index_mesh(df):
    ii1, ii2 = np.meshgrid(df.index, df.index)
    return pd.DataFrame({"row": ii1.ravel(), "col": ii2.ravel()})


def _elements_mesh(df, elem1, elem2):
    ee1, ee2 = np.meshgrid(df[elem1], df[elem2])
    return pd.DataFrame({"row": ee1.ravel(), "col": ee2.ravel()})


def edge_in_face_connectivity(eptm):
    """Returns an array of shape (eptm.Ne, eptm.Ne) with
    C_ij = 1 iff edges i and j belong to the same face.

    """
    mesh = eptm.edge_df.groupby("face").apply(_index_mesh)
    ef_connect = sparse.coo_matrix(
        (np.ones(mesh.shape[0]), (mesh["row"], mesh["col"])),
        shape=(eptm.Ne, eptm.Ne),
        dtype=int,
    ).toarray()
    return ef_connect


def face_face_connectivity(eptm, exclude_opposites=False):
    """Returns an array of shape (eptm.Nf, eptm.Nf) with
    C_ij = n, where n is the number of shared vertices
    between the faces i and j.

    Parameters
    ----------
    eptm: a :class:`tyssue.Epithelium` instance
    exclude_opposites: bool, default `False`
        if True, opposite faces are not included in the
        resulting connectivity matrix

    """
    mesh = eptm.edge_df.groupby("srce").apply(_elements_mesh, "face", "face")

    ff_connect = sparse.coo_matrix(
        (np.ones(mesh.shape[0]), (mesh["row"], mesh["col"])),
        shape=(eptm.Nf, eptm.Nf),
        dtype=int,
    ).toarray()

    ff_connect[np.arange(eptm.Nf), np.arange(eptm.Nf)] = 0
    if exclude_opposites:
        eptm.get_opposite_faces()
        oppos = eptm.face_df.query("opposite >= 0")["opposite"]
        ff_connect[oppos.index, oppos.values] = 0
    return ff_connect


def cell_cell_connectivity(eptm):
    """Returns an array of shape (eptm.Nc, eptm.Nc) with
    C_ij = n, where n is the number of connections
    between the cells i and j.

    """
    mesh = eptm.edge_df.groupby("srce").apply(_elements_mesh, "cell", "cell")

    cc_connect = sparse.coo_matrix(
        (np.ones(mesh.shape[0]), (mesh["row"], mesh["col"])),
        shape=(eptm.Nc, eptm.Nc),
        dtype=int,
    ).toarray()
    cc_connect[np.arange(eptm.Nc), np.arange(eptm.Nc)] = 0

    return cc_connect


def srce_trgt_connectivity(eptm):
    """Returns an array of shape (eptm.Nv, eptm.Nv) with
    C_ij = n, where n is the number of shared edges
    between the vertices i and j.

    """
    srce, trgt = eptm.edge_df["srce"], eptm.edge_df["trgt"]

    st_connect = sparse.coo_matrix(
        (np.ones(eptm.Ne), (srce, trgt)), shape=(eptm.Nv, eptm.Nv), dtype=int
    ).toarray()
    return st_connect


def verts_in_face_connectivity(eptm):
    """Returns an array of shape (eptm.Nv, eptm.Nv) with
    C_ij = n, where n is the number of shared faces
    between the vertices i and j.
    """
    mesh = eptm.edge_df.groupby("face").apply(_elements_mesh, "srce", "srce")

    fst_connect = sparse.coo_matrix(
        (np.ones(mesh.shape[0]), (mesh["row"], mesh["col"])),
        shape=(eptm.Nv, eptm.Nv),
        dtype=int,
    ).toarray()
    fst_connect[np.arange(eptm.Nv), np.arange(eptm.Nv)] = 0
    return fst_connect


def verts_in_cell_connectivity(eptm):
    """Returns an array of shape (eptm.Nv, eptm.Nv) with
    C_ij = n, where n is the number of shared cells
    between the vertices i and j.
    """
    mesh = eptm.edge_df.groupby("cell").apply(_elements_mesh, "srce", "srce")

    cst_connect = sparse.coo_matrix(
        (np.ones(mesh.shape[0]), (mesh["row"], mesh["col"])),
        shape=(eptm.Nv, eptm.Nv),
        dtype=int,
    ).toarray()
    cst_connect[np.arange(eptm.Nv), np.arange(eptm.Nv)] = 0
    return cst_connect



import logging
import warnings
import itertools
from functools import wraps

import numpy as np
import pandas as pd

from .sheet_topology import face_division
from .base_topology import (
    add_vert,
    close_face,
    condition_4i,
    condition_4ii,
    collapse_edge,
    remove_face,
)
from .base_topology import split_vert as base_split_vert
from ..geometry.utils import rotation_matrix
from ..core.monolayer import Monolayer
from ..core.sheet import get_opposite

logger = logging.getLogger(name=__name__)
MAX_ITER = 10


def check_condition4(func):
    @wraps(func)
    def decorated(eptm, *args, **kwargs):
        eptm.backup()
        res = func(eptm, *args, **kwargs)
        if len(condition_4i(eptm)) or len(condition_4ii(eptm)):
            print("Invalid epithelium produced, restoring")
            # print("4i on", condition_4i(eptm))
            # print("4ii on", condition_4ii(eptm))
            eptm.restore()
            eptm.topo_changed = True
        return res

    return decorated


def remove_cell(eptm, cell):
    """Removes a tetrahedral cell from the epithelium
    """
    eptm.get_opposite_faces()
    edges = eptm.edge_df.query(f"cell == {cell}")
    if not edges.shape[0] == 12:
        warnings.warn(f"{cell} is not a tetrahedral cell, aborting")
        return -1
    faces = eptm.face_df.loc[edges["face"].unique()]
    oppo = faces["opposite"][faces["opposite"] != -1]
    verts = eptm.vert_df.loc[edges["srce"].unique()].copy()

    eptm.vert_df = eptm.vert_df.append(verts.mean(), ignore_index=True)
    new_vert = eptm.vert_df.index[-1]

    eptm.vert_df.loc[new_vert, "segment"] = "basal"
    eptm.edge_df.replace(
        {"srce": verts.index, "trgt": verts.index}, new_vert, inplace=True
    )

    collapsed = eptm.edge_df.query("srce == trgt")

    eptm.face_df.drop(faces.index, axis=0, inplace=True)
    eptm.face_df.drop(oppo, axis=0, inplace=True)

    eptm.edge_df.drop(collapsed.index, axis=0, inplace=True)

    eptm.cell_df.drop(cell, axis=0, inplace=True)
    eptm.vert_df.drop(verts.index, axis=0, inplace=True)
    eptm.reset_index()
    eptm.reset_topo()
    assert eptm.validate()
    return 0


def close_cell(eptm, cell):
    """Closes the cell by adding a face. Assumes a single face is missing
    """
    eptm.face_df = eptm.face_df.append(eptm.face_df.iloc[0], ignore_index=True)

    new_face = eptm.face_df.index[-1]

    face_edges = eptm.edge_df[eptm.edge_df["cell"] == cell]
    oppo = get_opposite(face_edges)
    new_edges = face_edges[oppo == -1].copy()
    if not new_edges.shape[0]:
        return 0
    logger.info("closing cell %d", cell)
    new_edges[["srce", "trgt"]] = new_edges[["trgt", "srce"]]
    new_edges["face"] = new_face
    new_edges.index = new_edges.index + eptm.edge_df.index.max()
    eptm.edge_df = eptm.edge_df.append(new_edges, ignore_index=False)

    eptm.reset_index()
    eptm.reset_topo()
    return 0


def split_vert(eptm, vert, face=None, multiplier=1.5):
    """Splits a vertex towards a face.

    Parameters
    ----------
    eptm : a :class:`tyssue.Epithelium` instance
    vert : int the vertex to split
    face : int, optional, the face to split
        if face is None, one face will be chosen at random
    multiplier: float, default 1.5
        length of the new edge(s) in units of eptm.settings["threshold_length"]

    Note on the algorithm
    ---------------------

    For a given face, we look for the adjacent cell with the lowest number
    of faces converging on the vertex. If this number is higher than 4
    we raise a ValueError

    If it's 3, we do a OI transition, resulting in a new edge but no new faces
    If it's 4, we do a IH transition, resulting in a new face and 2 ne edges.

    see ../doc/illus/IH_transition.png
    """
    all_edges = eptm.edge_df[
        (eptm.edge_df["trgt"] == vert) | (eptm.edge_df["srce"] == vert)
    ]

    faces = all_edges.groupby("face").apply(
        lambda df: pd.Series(
            {
                "verts": frozenset(df[["srce", "trgt"]].values.ravel()),
                "cell": df["cell"].iloc[0],
            }
        )
    )

    cells = all_edges.groupby("cell").apply(
        lambda df: pd.Series(
            {
                "verts": frozenset(df[["srce", "trgt"]].values.ravel()),
                "faces": frozenset(df["face"]),
                "size": df.shape[0] // 2,
            }
        )
    )

    # choose a face
    if face is None:
        face = np.random.choice(faces.index)

    pair = faces[faces["verts"] == faces.loc[face, "verts"]].index
    # Take the cell adjacent to the face with the smallest size
    cell = cells.loc[faces.loc[pair, "cell"], "size"].idxmin()
    face = pair[0] if pair[0] in cells.loc[cell, "faces"] else pair[1]
    elements = vert, face, cell

    if cells.loc[cell, "size"] == 3:
        logger.info(f"OI for face {face} of cell {cell}")
        _OI_transition(eptm, all_edges, elements, multiplier)
    elif cells.loc[cell, "size"] == 4:
        logger.info(f"OH for face {face} of cell {cell}")
        _OH_transition(eptm, all_edges, elements, multiplier)
    else:
        raise ValueError(
            "Cell has too many edges connected to the vertex, try with another"
        )

    # Tidy up
    for face in all_edges["face"].unique():
        close_face(eptm, face)
    eptm.reset_index()
    eptm.reset_topo()

    for cell in all_edges["cell"]:
        close_cell(eptm, cell)

    eptm.reset_index()
    eptm.reset_topo()

    if isinstance(eptm, Monolayer):
        for vert_ in eptm.vert_df.index[-2:]:
            eptm.guess_vert_segment(vert_)
        for face_ in eptm.face_df.index[-2:]:
            eptm.guess_face_segment(face_)


def _OI_transition(eptm, all_edges, elements, multiplier=1.5, recenter=False):

    epsilon = eptm.settings.get("threshold_length", 0.1) * multiplier
    vert, face, cell = elements

    # Get all the edges bordering this terahedron
    cell_eges = eptm.edge_df.query(f"cell == {cell}")
    prev_vs = cell_eges[cell_eges["trgt"] == vert]["srce"]
    next_vs = cell_eges[cell_eges["srce"] == vert]["trgt"]

    connected = all_edges[
        all_edges["trgt"].isin(next_vs)
        | all_edges["srce"].isin(prev_vs)
        | all_edges["srce"].isin(next_vs)
        | all_edges["trgt"].isin(prev_vs)
    ]
    base_split_vert(eptm, vert, face, connected, epsilon, recenter)


def _OH_transition(eptm, all_edges, elements, multiplier=1.5, recenter=False):

    epsilon = eptm.settings.get("threshold_length", 0.1) * multiplier
    vert, face, cell = elements

    # all_cell_edges = eptm.edge_df.query(f'cell == {cell}').copy()
    cell_edges = all_edges.query(f"cell == {cell}").copy()

    face_verts = cell_edges.groupby("face").apply(
        lambda df: set(df["srce"]).union(df["trgt"]) - {vert}
    )

    for face_, verts_ in face_verts.items():
        if not verts_.intersection(face_verts.loc[face]):
            opp_face = face_
            break
    else:
        raise ValueError

    for to_split in (face, opp_face):
        face_edges = all_edges.query(f"face == {to_split}").copy()

        prev_v, = face_edges[face_edges["trgt"] == vert]["srce"]
        next_v, = face_edges[face_edges["srce"] == vert]["trgt"]
        connected = all_edges[
            all_edges["trgt"].isin((next_v, prev_v))
            | all_edges["srce"].isin((next_v, prev_v))
        ]
        base_split_vert(eptm, vert, to_split, connected, epsilon, recenter)


def get_division_edges(eptm, mother, plane_normal, plane_center=None):
    """Returns an index of the mother cell edges crossed by the division plane, ordered
    clockwize around the division plane normal.


    """
    plane_normal = np.asarray(plane_normal)
    if plane_center is None:
        plane_center = eptm.cell_df.loc[mother, eptm.coords]

    n_xy = np.linalg.norm(plane_normal[:2])
    theta = -np.arctan2(n_xy, plane_normal[2])
    if np.linalg.norm(plane_normal[:2]) < 1e-10:
        rot = None
    else:
        direction = [plane_normal[1], -plane_normal[0], 0]
        rot = rotation_matrix(theta, direction)
    cell_verts = set(eptm.edge_df[eptm.edge_df["cell"] == mother]["srce"])
    vert_pos = eptm.vert_df.loc[cell_verts, eptm.coords]
    for coord in eptm.coords:
        vert_pos[coord] -= plane_center[coord]
    if rot is not None:
        vert_pos[:] = np.dot(vert_pos, rot)

    mother_edges = eptm.edge_df[eptm.edge_df["cell"] == mother]
    srce_z = vert_pos.loc[mother_edges["srce"], "z"]
    srce_z.index = mother_edges.index
    trgt_z = vert_pos.loc[mother_edges["trgt"], "z"]
    trgt_z.index = mother_edges.index
    division_edges = mother_edges[((srce_z < 0) & (trgt_z >= 0))]

    # Order the returned edges so that their centers
    # are oriented counterclockwize in the division plane
    # in preparation for septum creation
    srce_pos = vert_pos.loc[division_edges["srce"], eptm.coords].values
    trgt_pos = vert_pos.loc[division_edges["trgt"], eptm.coords].values
    centers = (srce_pos + trgt_pos) / 2
    theta = np.arctan2(centers[:, 1], centers[:, 0])
    return division_edges.iloc[np.argsort(theta)].index


def get_division_vertices(
    eptm, division_edges=None, mother=None, plane_normal=None, plane_center=None
):

    if division_edges is None:
        division_edges = get_division_edges(eptm, mother, plane_normal, plane_center)
    vertices = []
    for edge in division_edges:
        new_vert, *_ = add_vert(eptm, edge)
        vertices.append(new_vert)
    return vertices


@check_condition4
def cell_division(eptm, mother, geom, vertices=None):

    if vertices is None:
        vertices = get_division_vertices(
            eptm,
            division_edges=None,
            mother=mother,
            plane_normal=None,
            plane_center=None,
        )

    cell_cols = eptm.cell_df.loc[mother]
    eptm.cell_df = eptm.cell_df.append(cell_cols, ignore_index=True)
    eptm.cell_df.index.name = "cell"
    daughter = eptm.cell_df.index[-1]

    pairs = set(
        [
            frozenset([v1, v2])
            for v1, v2 in itertools.product(vertices, vertices)
            if v1 != v2
        ]
    )
    daughter_faces = []

    # devide existing faces
    for v1, v2 in pairs:
        v1_faces = eptm.edge_df[eptm.edge_df["srce"] == v1]["face"]
        v2_faces = eptm.edge_df[eptm.edge_df["srce"] == v2]["face"]
        # we should devide a face if both v1 and v2
        # are part of it
        faces = set(v1_faces).intersection(v2_faces)
        for face in faces:
            daughter_faces.append(face_division(eptm, face, v1, v2))
    # septum
    face_cols = eptm.face_df.iloc[-2:]
    eptm.face_df = eptm.face_df.append(face_cols, ignore_index=True)
    eptm.face_df.index.name = "face"
    septum = eptm.face_df.index[-2:]
    daughter_faces.extend(list(septum))

    num_v = len(vertices)
    num_new_edges = num_v * 2

    edge_cols = eptm.edge_df.iloc[-num_new_edges:]
    eptm.edge_df = eptm.edge_df.append(edge_cols, ignore_index=True)
    eptm.edge_df.index.name = "edge"
    new_edges = eptm.edge_df.index[-num_new_edges:]

    # To keep mother orientation, the first septum face
    # belongs to mother
    for v1, v2, edge, oppo in zip(
        vertices, np.roll(vertices, -1), new_edges[:num_v], new_edges[num_v:]
    ):
        # Mother septum
        eptm.edge_df.loc[edge, ["srce", "trgt", "face", "cell"]] = (
            v1,
            v2,
            septum[0],
            mother,
        )
        # Daughter septum
        eptm.edge_df.loc[oppo, ["srce", "trgt", "face", "cell"]] = (
            v2,
            v1,
            septum[1],
            daughter,
        )

    eptm.reset_index()
    eptm.reset_topo()
    geom.update_all(eptm)

    m_septum_edges = eptm.edge_df[eptm.edge_df["face"] == septum[0]]
    m_septum_norm = m_septum_edges[eptm.ncoords].mean()
    m_septum_pos = eptm.face_df.loc[septum[0], eptm.coords]

    # splitting the faces between mother and daughter
    # based on the orientation of the vector from septum
    # center to each face center w/r to the septum norm
    mother_faces = set(eptm.edge_df[eptm.edge_df["cell"] == mother]["face"])
    for face in mother_faces:
        if face == septum[0]:
            continue

        dr = eptm.face_df.loc[face, eptm.coords] - m_septum_pos
        proj = (dr.values * m_septum_norm).sum(axis=0)
        f_edges = eptm.edge_df[eptm.edge_df["face"] == face].index
        if proj < 0:
            eptm.edge_df.loc[f_edges, "cell"] = mother
        else:
            eptm.edge_df.loc[f_edges, "cell"] = daughter
    eptm.reset_index()
    eptm.reset_topo()
    return daughter


def find_rearangements(eptm):
    """Finds the candidates for IH and HI transitions
    Returns
    -------
    edges_HI: set of indexes of short edges
    faces_IH: set of indexes of small triangular faces
    """
    l_th = eptm.settings.get("threshold_length", 1e-6)
    shorts = eptm.edge_df[eptm.edge_df["length"] < l_th]
    if not shorts.shape[0]:
        return np.array([]), np.array([])
    edges_IH = find_IHs(eptm, shorts)
    faces_HI = find_HIs(eptm, shorts)
    return edges_IH, faces_HI


def find_IHs(eptm, shorts=None):

    l_th = eptm.settings.get("threshold_length", 1e-6)
    if shorts is None:
        shorts = eptm.edge_df[eptm.edge_df["length"] < l_th]
    if not shorts.shape[0]:
        return []

    edges_IH = shorts.groupby("srce").apply(
        lambda df: pd.Series(
            {
                "edge": df.index[0],
                "length": df["length"].iloc[0],
                "num_sides": min(eptm.face_df.loc[df["face"], "num_sides"]),
                "pair": frozenset(df.iloc[0][["srce", "trgt"]]),
            }
        )
    )
    # keep only one of the edges per vertex pair and sort by length
    edges_IH = (
        edges_IH[edges_IH["num_sides"] > 3]
        .drop_duplicates("pair")
        .sort_values("length")
    )
    return edges_IH["edge"].values


def find_HIs(eptm, shorts=None):
    l_th = eptm.settings.get("threshold_length", 1e-6)
    if shorts is None:
        shorts = eptm.edge_df[(eptm.edge_df["length"] < l_th)]
    if not shorts.shape[0]:
        return []

    max_f_length = shorts.groupby("face")["length"].apply(max)
    short_faces = eptm.face_df.loc[max_f_length[max_f_length < l_th].index]
    faces_HI = short_faces[short_faces["num_sides"] == 3].sort_values("area").index
    return faces_HI


@check_condition4
def IH_transition(eptm, edge):
    """
    I → H transition as defined in Okuda et al. 2013
    (DOI 10.1007/s10237-012-0430-7).
    See tyssue/doc/illus/IH_transition.png for the algorithm
    """
    srce, trgt, face, cell = eptm.edge_df.loc[edge, ["srce", "trgt", "face", "cell"]]
    vert = min(srce, trgt)
    collapse_edge(eptm, edge)

    split_vert(eptm, vert, face)

    logger.info(f"IH transition on edge {edge}")
    return 0


@check_condition4
def HI_transition(eptm, face):
    """
    H → I transition as defined in Okuda et al. 2013
    (DOI 10.1007/s10237-012-0430-7).
    See tyssue/doc/illus/IH_transition.png for the algorithm
    """
    remove_face(eptm, face)
    vert = eptm.vert_df.index[-1]
    all_edges = eptm.edge_df[
        (eptm.edge_df["srce"] == vert) | (eptm.edge_df["trgt"] == vert)
    ]

    cells = all_edges.groupby("cell").size()
    cell = cells.idxmin()
    face = all_edges[all_edges["cell"] == cell]["face"].iloc[0]
    split_vert(eptm, vert, face)

    logger.info(f"HI transition on edge {face}")
    return 0


def _add_edge_to_existing(eptm, cell, vi, vj, new_srce, new_trgt):
    """
    Add edges between vertices v7, v8 and v9 to the existing faces
    """
    cell_edges = eptm.edge_df[eptm.edge_df["cell"] == cell]
    for f, data in cell_edges.groupby("face"):
        if {vi, vj, new_srce, new_trgt}.issubset(set(data["srce"]).union(data["trgt"])):
            good_f = f
            break
    else:
        raise ValueError(
            "no face with vertices {}, {}, {} and {}"
            " was found for cell {}".format(vi, vj, new_srce, new_trgt, cell)
        )
    eptm.edge_df = eptm.edge_df.append(cell_edges.iloc[-1], ignore_index=True)
    new_e = eptm.edge_df.index[-1]
    eptm.edge_df.loc[new_e, ["srce", "trgt", "face", "cell"]] = (
        new_srce,
        new_trgt,
        good_f,
        cell,
    )


def _set_new_pos_IH(eptm, e_1011, vertices):
    """Okuda 2013 equations 46 to 56
    """
    Dl_th = eptm.settings["threshold_length"]

    (v1, v2, v3, v4, v5, v6, v7, v8, v9, v10, v11) = vertices

    # eq. 49
    r_1011 = -eptm.edge_df.loc[e_1011, eptm.dcoords].values
    u_T = r_1011 / np.linalg.norm(r_1011)
    # eq. 50
    r0 = eptm.vert_df.loc[[v10, v11], eptm.coords].mean(axis=0).values

    v_0ns = []
    for vi, vj, vk in zip((v1, v2, v3), (v4, v5, v6), (v7, v8, v9)):
        # eq. 54 - 56
        r0i, r0j = eptm.vert_df.loc[[vi, vj], eptm.coords].values - r0[np.newaxis, :]
        w_0k = (r0i / np.linalg.norm(r0i) + r0j / np.linalg.norm(r0j)) / 2
        # eq. 51 - 53
        v_0k = w_0k - (np.dot(w_0k, u_T)) * u_T
        v_0ns.append(v_0k)

    # see definition of l_max bellow eq. 56
    l_max = np.max(
        [np.linalg.norm(v_n - v_m) for (v_n, v_m) in itertools.combinations(v_0ns, 2)]
    )
    # eq. 46 - 49
    for vk, v_0k in zip((v7, v8, v9), v_0ns):
        eptm.vert_df.loc[vk, eptm.coords] = r0 + (Dl_th / l_max) * v_0k


def _get_vertex_pairs_IH(eptm, e_1011):

    srce_face_orbits = eptm.get_orbits("srce", "face")
    v10, v11 = eptm.edge_df.loc[e_1011, ["srce", "trgt"]]
    common_faces = set(srce_face_orbits.loc[v10]).intersection(
        srce_face_orbits.loc[v11]
    )
    if eptm.face_df.loc[common_faces, "num_sides"].min() < 4:
        logger.warning(
            "Edge %i has adjacent triangular faces"
            " can't perform IH transition, aborting",
            e_1011,
        )
        return None

    v10_out = set(eptm.edge_df[eptm.edge_df["srce"] == v10]["trgt"]) - {v11}
    faces_123 = {
        v: set(srce_face_orbits.loc[v])  # .intersection(srce_face_orbits.loc[v10])
        for v in v10_out
    }

    v11_out = set(eptm.edge_df[eptm.edge_df["srce"] == v11]["trgt"]) - {v10}
    faces_456 = {
        v: set(srce_face_orbits.loc[v])  # .intersection(srce_face_orbits.loc[v11])
        for v in v11_out
    }
    v_pairs = []
    for vi in v10_out:
        for vj in v11_out:
            common_face = faces_123[vi].intersection(faces_456[vj])
            if common_face:
                v_pairs.append((vi, vj))
                break
        else:
            return None
    return v_pairs


def _set_new_pos_HI(eptm, fa, v10, v11):

    r0 = eptm.face_df.loc[fa, eptm.coords].values

    norm_a = eptm.edge_df[eptm.edge_df["face"] == fa][eptm.ncoords].mean(axis=0).values
    norm_a = norm_a / np.linalg.norm(norm_a)
    norm_b = -norm_a
    Dl_th = eptm.settings["threshold_length"] * 1.01
    eptm.vert_df.loc[v10, eptm.coords] = r0 + Dl_th / 2 * norm_b
    eptm.vert_df.loc[v11, eptm.coords] = r0 + Dl_th / 2 * norm_a



import numpy as np
from .base_geometry import BaseGeometry


class PlanarGeometry(BaseGeometry):
    """Geomtetry methods for 2D planar cell arangements
    """

    @classmethod
    def update_all(cls, sheet):
        """
        Updates the sheet geometry by updating:
        * the edge vector coordinates
        * the edge lengths
        * the face centroids
        * the normals to each edge associated face
        * the face areas
        """

        cls.update_dcoords(sheet)
        cls.update_length(sheet)
        cls.update_centroid(sheet)
        cls.update_normals(sheet)
        cls.update_areas(sheet)
        cls.update_perimeters(sheet)

    @staticmethod
    def update_normals(sheet):

        coords = sheet.coords
        face_pos = sheet.edge_df[["f" + c for c in coords]].values
        srce_pos = sheet.edge_df[["s" + c for c in coords]].values
        trgt_pos = sheet.edge_df[["t" + c for c in coords]].values
        normals = np.cross(srce_pos - face_pos, trgt_pos - srce_pos)
        sheet.edge_df["nz"] = normals

    @staticmethod
    def update_areas(sheet):
        """
        Updates the normal coordinate of each (srce, trgt, face) face.
        """
        sheet.edge_df["sub_area"] = sheet.edge_df["nz"] / 2
        sheet.face_df["area"] = sheet.sum_face(sheet.edge_df["sub_area"])

    @staticmethod
    def face_projected_pos(sheet, face, psi):
        """
        returns the sheet vertices position translated to center the face
        `face` at (0, 0) and rotated in the (x, y) plane
        by and angle `psi` radians

        """
        rot_pos = sheet.vert_df[sheet.coords].copy()
        face_x, face_y = sheet.face_df.loc[face, ["x", "y"]]
        rot_pos.x = (sheet.vert_df.x - face_x) * np.cos(psi) - (
            sheet.vert_df.y - face_y
        ) * np.sin(psi)
        rot_pos.y = (sheet.vert_df.x - face_x) * np.sin(psi) + (
            sheet.vert_df.y - face_y
        ) * np.cos(psi)

        return rot_pos


# The following classes will probably be included in tyssue at some point
class AnnularGeometry(PlanarGeometry):
    """
    """

    @classmethod
    def update_all(cls, eptm):
        PlanarGeometry.update_all(eptm)
        cls.update_lumen_volume(eptm)

    @staticmethod
    def update_lumen_volume(eptm):
        srce_pos = eptm.upcast_srce(eptm.vert_df[["x", "y"]]).loc[eptm.apical_edges]
        trgt_pos = eptm.upcast_trgt(eptm.vert_df[["x", "y"]]).loc[eptm.apical_edges]
        apical_edge_pos = (srce_pos + trgt_pos) / 2
        apical_edge_coords = eptm.edge_df.loc[eptm.apical_edges, ["dx", "dy"]]
        eptm.settings["lumen_volume"] = (
            -apical_edge_pos["x"] * apical_edge_coords["dy"]
            + apical_edge_pos["y"] * apical_edge_coords["dx"]
        ).values.sum()



import warnings
from .plt_draw import quick_edge_draw, plot_forces
from .plt_draw import sheet_view as sheet_view_2d
from .ipv_draw import browse_history

from .plt_draw import sheet_view as sheet_view_2d
from .ipv_draw import sheet_view as sheet_view_3d
from .plt_draw import create_gif


def sheet_view(sheet, coords=["x", "y", "z"], ax=None, mode="2D", **draw_specs_kw):
    """Main plotting function in 2D or 3D.


    Parameters
    ----------
    sheet: :class:`Epithelium` instance
    coords: list of strings,
     the coordinates over which to do the plot
    ax: :class:matplotlib.Axes instance, default None
     axis over which to plot the sheet, for quick and
    mode: str, {'2D'|'quick'|'3D'}, default '2D'
     the type of graph to plot (see bellow)

    Returns
    -------
    fig, {ax|meshes}:


    """

    if mode == "2D":
        return sheet_view_2d(sheet, coords[:2], ax, **draw_specs_kw)
    elif mode == "quick":
        edge_kw = draw_specs_kw.get("edge", {})
        return quick_edge_draw(sheet, coords[:2], ax, **edge_kw)
    elif mode == "3D":
        return sheet_view_3d(sheet, coords, **draw_specs_kw)

    return ValueError(
        """
Argument `mode` not understood,
should be either '2D', '3D' or 'quick', got %s""",
        mode,
    )


def highlight_cells(eptm, cells, reset_visible=False):
    """Sets a column 'visible' to True for all the faces of the
    cells passed as argument (for a 3D tyssue).

    If no such column exists in eptm.face_df, creates it.

    """
    if reset_visible:
        eptm.face_df["visible"] = False

    if not hasattr(cells, "__iter__"):
        cells = [cells]

    for cell in cells:
        cell_faces = eptm.edge_df[eptm.edge_df["cell"] == cell]["face"]
        highlight_faces(eptm.face_df, cell_faces, reset_visible=False)


def highlight_faces(face_df, faces, reset_visible=False):
    """
    Sets the faces visibility to 1

    If `reset_visible` is `True`, sets all the other faces
    to `visible = False`
    """
    if ("visible" not in face_df.columns) or reset_visible:
        face_df["visible"] = False

    face_df.loc[faces, "visible"] = True



"""
Apoptosis event module
=======================


"""

from ...utils.decorators import face_lookup
from ...geometry.sheet_geometry import SheetGeometry

from .actions import shrink, ab_pull, exchange, remove
from .basic_events import contraction

default_apoptosis_spec = {
    "face_id": -1,
    "face": -1,
    "shrink_rate": 0.1,
    "critical_area": 1e-2,
    "radial_tension": 0.1,
    "contractile_increase": 0.1,
    "contract_span": 2,
    "geom": SheetGeometry,
}


@face_lookup
def apoptosis(sheet, manager, **kwargs):
    """Apoptotic behavior

    While the cell's apical area is bigger than a threshold, the
    cell shrinks, and the contractility of its neighbors is increased.
    once the critical area is reached, the cell is eliminated
    from the apical surface through successive type 1 transition. Once
    only three sides are left, the cell is eliminated from the tissue.

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    manager : a :class:`EventManager` object
    face_id : int,
        the id of the apoptotic cell
    shrink_rate : float, default 0.1
        the rate of reduction of the cell's prefered volume
        e.g. the prefered volume is devided by a factor 1+shrink_rate
    critical_area : area at which the face is eliminated from the sheet
    radial_tension : amount of radial tension added at each contraction steps
    contractile_increase : increase in contractility at the cell neighbors
    contract_span : number of neighbors affected by the contracitity increase
    geom : the geometry class used
    """

    apoptosis_spec = default_apoptosis_spec
    apoptosis_spec.update(**kwargs)
    face = apoptosis_spec["face"]

    if sheet.face_df.loc[face, "area"] > apoptosis_spec["critical_area"]:
        # Shrink and pull
        shrink(sheet, face, apoptosis_spec["shrink_rate"])
        ab_pull(sheet, face, apoptosis_spec["radial_tension"])
        # contract neighbors
        neighbors = sheet.get_neighborhood(
            face, apoptosis_spec["contract_span"]
        ).dropna()
        neighbors["id"] = sheet.face_df.loc[neighbors.face, "id"].values
        manager.extend(
            [
                (
                    contraction,
                    {
                        "face_id": neighbor["id"],
                        "contractile_increase": (
                            apoptosis_spec["contractile_increase"] / neighbor["order"],
                        ),
                    },
                )
                for _, neighbor in neighbors.iterrows()
            ]
        )
        done = False
    else:
        if sheet.face_df.loc[face, "num_sides"] > 3:
            exchange(sheet, face, apoptosis_spec["geom"])
            done = False
        else:
            remove(sheet, face, apoptosis_spec["geom"])
            done = True
    if not done:
        manager.append(apoptosis, **apoptosis_spec)



from ..json_parser import load_spec
import os

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))


def sheet_spec():
    """Default specification for drawing/graphical output
    functions for a sheet - also suitable for general faceted tissue

    .. code-block::

        {
            "edge": {
            "visible": true,
                "width": 0.5,
                "head_width": 0.2,
                "length_includes_head": true,
                "shape": "right",
                "color": "#2b5d0a",
                "alpha": 0.8,
                "zorder": 1
            },
            "vert": {
                "visible": true,
                "s": 100,
                "color": "#000a4b",
                "alpha": 0.3,
                "zorder": 2
            },
            "grad": {
                "color":"#000a4b",
                "alpha":0.5,
                "width":0.04
            },
            "face": {
                "visible": false,
                "color":"#8aa678",
                "alpha": 1.0,
                "zorder": -1
            }
        }

    """
    specfile = os.path.join(CURRENT_DIR, "sheet.json")
    return load_spec(specfile)



"""
Base gradients for sheet like geometries
"""

import numpy as np
import pandas as pd

from ..utils.utils import to_nd  # _to_3d, _to_2d


def height_grad(sheet):

    w = sheet.settings.get("height_axis", "z")
    u, v = (c for c in sheet.coords if c != w)
    coords = [u, v, w]

    if sheet.settings.get("geometry", "cylindrical") == "cylindrical":
        r_to_rho = sheet.vert_df[coords] / to_nd(sheet.vert_df["rho"], 3)
        r_to_rho[w] = 0.0

    elif sheet.settings.get("geometry") == "flat":
        r_to_rho = sheet.vert_df[coords].copy()
        r_to_rho[[u, v]] = 0.0
        r_to_rho[[w]] = 1.0

    elif sheet.settings.get("geometry") == "spherical":
        r_to_rho = sheet.vert_df[coords] / to_nd(sheet.vert_df["rho"], 3)

    return r_to_rho


def area_grad(sheet):

    coords = sheet.coords
    ncoords = sheet.ncoords
    inv_area = sheet.edge_df.eval("1 / (4 * sub_area)")
    # Some segmentations create null areas
    inv_area.replace(np.inf, 0, inplace=True)
    inv_area.replace(-np.inf, 0, inplace=True)

    face_pos = sheet.edge_df[["f" + c for c in coords]].values
    srce_pos = sheet.edge_df[["s" + c for c in coords]].values
    trgt_pos = sheet.edge_df[["t" + c for c in coords]].values

    r_ak = srce_pos - face_pos
    r_aj = trgt_pos - face_pos

    inv_area = to_nd(inv_area, 3)
    grad_a_srce = inv_area * np.cross(r_aj, sheet.edge_df[ncoords])
    grad_a_trgt = inv_area * np.cross(sheet.edge_df[ncoords], r_ak)
    return (
        pd.DataFrame(grad_a_srce, index=sheet.edge_df.index, columns=sheet.coords),
        pd.DataFrame(grad_a_trgt, index=sheet.edge_df.index, columns=sheet.coords),
    )



import os
import pandas as pd
import logging

logger = logging.getLogger(name=__name__)


def load_datasets(h5store, data_names=["face", "vert", "edge", "cell", "settings"]):
    if not os.path.isfile(h5store):
        raise FileNotFoundError("file %s not found" % h5store)
    with pd.HDFStore(h5store) as store:
        data = {name: store[name] for name in data_names if name in store}
    return data


def save_datasets(h5store, eptm):
    with pd.HDFStore(h5store) as store:
        for key in eptm.data_names:
            store.put(key, getattr(eptm, "{}_df".format(key)))



import time

from functools import wraps


def do_undo(func):
    """Decorator that creates a copy of the first argument
    (usually an epithelium object) and restores it if the function fails.

    The first argument in `*args` should have `backup()` and `restore()` methods.
    """

    @wraps(func)
    def with_bckup(*args, **kwargs):
        eptm = args[0]
        eptm.backup()
        try:
            res = func(*args, **kwargs)
            return res
        except Exception as err:
            eptm.restore()
            raise err

    return with_bckup


def validate(func):
    """Decorator that validate the epithelium after the
    decorated function was applied. the first argument
    of `func` should be an epithelium instance, and
    is at least assumed to have a `validate` method.
    """

    @wraps(func)
    def with_validate(*args, **kwargs):
        eptm = args[0]
        result = func(*args, **kwargs)
        if not eptm.validate():
            raise ValueError(
                """
An invalid epithelium was produced

To see which edges are invalid, you can inspect
the 'is_valid' column of the `edge_df` dataframe,
or for example the bad cells involved:

>>> bad_edges = eptm.edge_df[~eptm.edge_df['is_valid']].index
>>> bad_cells = eptm.edge_df.loc[bad_edges, 'cell'].unique()

If case the epithelium was restored after being invalidated, you can find the
invalid epithelium as the `_bad` attribute of the restored one"""
            )

        return result

    return with_validate


def time_exe(func):
    def with_time_exe(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()

        print("function : {} \ttime: {2:2f}sec".format(func.__name__, end - start))

        return result

    return with_time_exe


def face_lookup(func):
    @wraps(func)
    def with_face_lookup(*args, **kwargs):
        sheet = args[0]
        face_id = kwargs["face_id"]
        face = sheet.idx_lookup(face_id, "face")
        if face is None:
            return
        kwargs["face"] = face
        return func(*args, **kwargs)

    return with_face_lookup


def cell_lookup(func):
    @wraps(func)
    def with_cell_lookup(*args, **kwargs):
        sheet = args[0]
        cell_id = kwargs["cell_id"]
        cell = sheet.idx_lookup(cell_id, "cell")
        if cell is None:
            return
        kwargs["cell"] = cell
        return func(*args, **kwargs)

    return with_cell_lookup



def planar_sheet():
    """
    Specification for a 2D sheet in the plane
    """
    spec = {
        "edge": {
            "trgt": 0,
            "nz": 1.0,
            "length": 1.0,
            "face": 0,
            "srce": 0,
            "dx": 0.0,
            "dy": 0.0,
            "sx": 0.0,
            "sy": 0.0,
            "tx": 0.0,
            "ty": 0.0,
            "fx": 0.0,
            "fy": 0.0,
        },
        "vert": {"y": 0.0, "is_active": 1, "x": 0.0},
        "face": {
            "y": 0.0,
            "is_alive": 1,
            "perimeter": 0.0,
            "area": 0.0,
            "x": 0.0,
            "num_sides": 6,
        },
        "settings": {"geometry": "planar"},
    }
    return spec


def planar_spec():
    return planar_sheet()


def planar_periodic_sheet():
    spec = planar_sheet()
    spec["settings"] = {"boundaries": {"x": [-1, 1], "y": [-1, 1]}}
    return spec


def sheet_spec():

    spec = {
        "face": {
            "x": 0.0,
            "y": 0.0,
            "z": 0.0,
            "num_sides": 6,
            "area": 1.0,
            "perimeter": 1.0,
            "is_alive": 1,
        },
        "vert": {
            "x": 0.0,
            "y": 0.0,
            "z": 0.0,
            "is_active": 1,
            "rho": 0.0,
            "height": 0.0,
            "basal_shift": 4.0,
        },
        "edge": {
            "srce": 0,
            "trgt": 0,
            "face": 0,
            "dx": 0.0,
            "dy": 0.0,
            "dz": 0.0,
            "nx": 0.0,
            "ny": 0.0,
            "nz": 1.0,
            "sx": 0.0,
            "sy": 0.0,
            "sz": 0.0,
            "tx": 0.0,
            "ty": 0.0,
            "tz": 0.0,
            "fx": 0.0,
            "fy": 0.0,
            "fz": 0.0,
            "length": 1.0,
            "is_active": 1,
        },
        "settings": {"geometry": "cylindrical", "height_axis": "z"},
    }
    return spec


def periodic_sheet():
    spec = flat_sheet()
    spec["settings"].update({"boundaries": {"x": [-1, 1], "y": [-1, 1]}})
    return spec


def cylindrical_sheet():
    """Geometry specification of a sheet in a 3D space
    """
    spec = sheet_spec()
    return spec


def rod_sheet():
    """Geomtetry specs of a rod sheet in 3D space

    """
    spec = sheet_spec()
    spec["settings"].update({"geometry": "rod", "height_axis": "z", "ab": [0.0, 0.0]})
    spec["vert"].update({"left_tip": False, "right_tip": False})

    return spec


def flat_sheet():
    """Geometry specification of a sheet in a 3D space

    """
    spec = sheet_spec()
    spec["settings"].update({"geometry": "flat"})
    return spec


def spherical_sheet():
    """Geometry specification of a sheet in a 3D space.

    Height is computed with respect to the distance to
    the coordinate systems center

    """

    spec = sheet_spec()
    spec["settings"].update({"geometry": "spherical"})
    return spec


def bulk_spec():
    """ Geometry specification for bulk tissues

    """
    spec = sheet_spec()

    spec["edge"].update({"cx": 0.0, "cy": 0.0, "cz": 0.0, "cell": 0, "sub_vol": 0.0})

    spec["cell"] = {
        "x": 0.0,
        "y": 0.0,
        "z": 0.0,
        "area": 0.0,
        "vol": 0.0,
        "num_faces": 6,
        "is_alive": 1,
    }
    return spec



import numpy as np
import pandas as pd
from scipy.interpolate import Rbf

from .sheet import Sheet


class MultiSheet:
    def __init__(self, name, layer_datasets, specs):

        self.coords = ["x", "y", "z"]
        self.layers = [
            Sheet("{}_{}".format(name, i), dsets, specs, coords=self.coords)
            for i, dsets in enumerate(layer_datasets)
        ]
        for i, layer in enumerate(self):
            for dset in layer.datasets.values():
                dset["layer"] = i

    def __iter__(self):
        for layer in self.layers:
            yield layer

    def __getitem__(self, n):
        return self.layers[n]

    def __len__(self):
        return len(self.layers)

    @property
    def Nes(self):
        return [layer.Ne for i, layer in self]

    @property
    def Nvs(self):
        return [layer.Nv for i, layer in self]

    @property
    def Nfs(self):
        return [layer.Nf for i, layer in self]

    @property
    def v_idxs(self):
        return np.array([sheet.Nv for sheet in self]).cumsum()

    @property
    def f_idxs(self):
        return np.array([sheet.Nf for sheet in self]).cumsum()

    @property
    def e_idxs(self):
        return np.array([sheet.Ne for sheet in self]).cumsum()

    def concat_datasets(self):
        datasets = {}

        v_dfs = [self[0].vert_df]
        e_dfs = [self[0].edge_df]
        f_dfs = [self[0].face_df]

        v_shift = 0
        f_shift = 0
        e_shift = 0
        for lower, upper in zip(self[:-1], self[1:]):
            v_shift += lower.Nv
            v_dfs.append(upper.vert_df.set_index(upper.vert_df.index + v_shift))
            f_shift += lower.Nf
            f_dfs.append(upper.face_df.set_index(upper.face_df.index + f_shift))
            e_shift += lower.Ne
            shifted_edge_df = upper.edge_df.set_index(upper.edge_df.index + e_shift)
            shifted_edge_df[["srce", "trgt"]] += v_shift
            shifted_edge_df["face"] += f_shift
            e_dfs.append(shifted_edge_df)

        for key, dfs in zip(["edge", "face", "vert"], [e_dfs, f_dfs, v_dfs]):
            datasets[key] = pd.concat(dfs)
        return datasets

    def update_interpolants(self):

        self.interpolants = [
            Rbf(
                sheet.vert_df["x"],
                sheet.vert_df["y"],
                sheet.vert_df["z"],
                **sheet.specs["settings"]["interpolate"]
            )
            for sheet in self
        ]
        # for interp in self.interpolants:
        #     interp.nodes = interp.nodes.clip(-1e2, 1e2)



from .sheet_geometry import SheetGeometry


class MultiSheetGeometry:
    """ Geometry class for stacked 2D sheets

    """

    @classmethod
    def update_all(cls, msheet):

        msheet.update_interpolants()
        for sheet in msheet:
            SheetGeometry.update_dcoords(sheet)
            SheetGeometry.update_length(sheet)
            SheetGeometry.update_centroid(sheet)
            SheetGeometry.update_normals(sheet)
            SheetGeometry.update_areas(sheet)
            SheetGeometry.update_perimeters(sheet)
        cls.update_heights(msheet)

    @staticmethod
    def update_heights(msheet):

        msheet[0].vert_df["height"] = (
            msheet[0].vert_df["z"] - msheet[0].vert_df["basal_shift"]
        )

        #  Here we use basal_shift to impose apical constrains
        msheet[-1].vert_df["depth"] = (
            msheet[-1].vert_df["basal_shift"] - msheet[-1].vert_df["z"]
        )

        for lower, upper in zip(msheet.interpolants[:-1], msheet[1:]):
            upper.vert_df["height"] = upper.vert_df["z"] - lower(
                upper.vert_df["x"], upper.vert_df["y"]
            )
        for lower, upper in zip(msheet[:-1], msheet.interpolants[1:]):
            lower.vert_df["depth"] = (
                upper(lower.vert_df["x"], lower.vert_df["y"]) - lower.vert_df["z"]
            )



import os
import re
import sys
import platform
import subprocess
import warnings

from distutils.version import LooseVersion
from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext


DISTNAME = "tyssue"
DESCRIPTION = "tyssue is a living tissues, cell level, modeling library"
LONG_DESCRIPTION = (
    "tyssue uses the scientific python ecosystem to model"
    " epithelium at the cellular level"
)
MAINTAINER = "Guillaume Gay"
MAINTAINER_EMAIL = "guillaume@damcb.com"
URL = "https://github.com/DamCB/tyssue"
LICENSE = "GPL-3.0"
DOWNLOAD_URL = "https://github.com/DamCB/tyssue.git"

files = ["*.so*", "*.a*", "*.lib*", "config/*/*.json", "stores/*.*"]


## Version management copied form numpy
## Thanks to them!
MAJOR = 0
MINOR = 6
MICRO = 7
ISRELEASED = True
VERSION = "%d.%d.%s" % (MAJOR, MINOR, MICRO)


def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ["SYSTEMROOT", "PATH"]:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env["LANGUAGE"] = "C"
        env["LANG"] = "C"
        env["LC_ALL"] = "C"
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(["git", "rev-parse", "HEAD"])
        GIT_REVISION = out.strip().decode("ascii")
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


def get_version_info():
    # Adding the git rev number needs to be done inside write_version_py(),
    # otherwise the import of tyssue.version messes up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists(".git"):
        GIT_REVISION = git_version()
    elif os.path.exists("tyssue/version.py"):
        # must be a source distribution, use existing version file
        # read from it instead of importing to avoid importing
        # the whole package
        with open("tyssue/version.py", "r") as fh:
            for line in fh.readlines():
                if line.startswith("git_revision"):
                    GIT_REVISION = line.split("=")[-1][2:-2]
                    break
            else:
                GIT_REVISION = "Unknown"
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += ".dev0+" + GIT_REVISION[:7]

    return FULLVERSION, GIT_REVISION


def write_version_py(filename="tyssue/version.py"):
    fullversion, git_revision = get_version_info()

    with open(filename, "w") as a:
        a.write(
            f"""
# THIS FILE IS GENERATED FROM tyssue SETUP.PY
#
short_version = '{VERSION}'
full_version = '{fullversion}'
git_revision = '{git_revision}'
release = {ISRELEASED}
if release:
    version = full_version
else:
    version = short_version
"""
        )


write_version_py()
setup(
    name=DISTNAME,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    url=URL,
    license=LICENSE,
    download_url=DOWNLOAD_URL,
    version=VERSION,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    packages=find_packages(),
    package_data={"tyssue": files},
    include_package_data=True,
    zip_safe=False,
)



import logging
from ..topology import TopologyChangeError

log = logging.getLogger(__name__)


def set_pos(eptm, geom, pos):
    """Updates the vertex position of the :class:`Epithelium` object.

    Assumes that pos is passed as a 1D array to be reshaped as (eptm.Nv, eptm.dim)

    """
    log.debug("set pos")
    if eptm.topo_changed:
        # reset the switch and interupt what we were doing
        eptm.topo_changed = False
        raise TopologyChangeError
    eptm.vert_df.loc[eptm.active_verts, eptm.coords] = pos.reshape((-1, eptm.dim))
    geom.update_all(eptm)



"""
Basic event module
=======================


"""

import logging
import numpy as np
from ...topology.base_topology import collapse_edge
from ...topology.sheet_topology import remove_face, type1_transition
from ...topology.sheet_topology import split_vert as sheet_split
from ...topology.bulk_topology import split_vert as bulk_split

from ...geometry.sheet_geometry import SheetGeometry
from ...core.sheet import Sheet
from ...utils import connectivity

import warnings

logger = logging.getLogger(__name__)


def merge_vertices(sheet):
    """Merges all the vertices that are closer than the threshold length

    Parameters
    ----------
    sheet : a :class:`Sheet` object

    """
    d_min = sheet.settings.get("threshold_length", 1e-3)
    short = sheet.edge_df[sheet.edge_df["length"] < d_min].index
    if not short.shape[0]:
        return -1
    logger.info(f"Collapsing {short.shape[0]} edges")
    while short.shape[0]:
        collapse_edge(sheet, short[0], allow_two_sided=True)
        short = sheet.edge_df[sheet.edge_df["length"] < d_min].index
    return 0


def detach_vertices(sheet):
    """Stochastically detaches vertices from rosettes.


    Uses two probabilities `p_4` and `p_5p` stored in
    sheet.settings.

    Parameters
    ----------
    sheet : a :class:`Sheet` object

    """
    # sheet.update_rank()
    st_connect = connectivity.srce_trgt_connectivity(sheet)
    rank = ((st_connect + st_connect.T) > 0).sum(axis=0)
    if isinstance(sheet, Sheet):
        min_rank = 3
        split_vert = sheet_split
    else:
        min_rank = 4
        split_vert = bulk_split

    if rank.max() == min_rank:
        return 0

    dt = sheet.settings.get("dt", 1.0)
    p_4 = sheet.settings.get("p_4", 0.1) * dt
    p_5p = sheet.settings.get("p_5p", 1e-2) * dt

    rank4 = sheet.vert_df[rank == min_rank + 1].index
    dice4 = np.random.random(rank4.size)

    rank5p = sheet.vert_df[rank > min_rank + 1].index
    dice5p = np.random.random(rank5p.size)

    to_detach = np.concatenate([rank4[dice4 < p_4], rank5p[dice5p < p_5p]])
    logger.info(f"Detaching {to_detach.size} vertices")

    for vert in to_detach:
        split_vert(sheet, vert)


def set_value(sheet, element, index, set_value, col):
    """Set the value in the dataset at position index/col.

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    element : str: 'cell' or 'face' or 'edge' or 'vert'
    index : index in the datasets[element]
    set_value : value to set.
    col : column from dataset which apply increase_rate.
    """
    sheet.datasets[element].loc[index, col] = set_value


def increase(sheet, element, index, increase_rate, col, multiply=True, bound=None):
    """Increase the value in the dataset at position index/col.

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    element : str: 'cell' or 'face' or 'edge' or 'vert'
    index : index in the datasets[element]
    increase_rate : rate use to multiply value in the column col.
    col : column from dataset which apply increase_rate.
    multiply : bool: if true, the current col value is multiplied by increase_rate. if false it is added. Default multiply.
    bound: Higher limit of the modify value. Default None
    """
    if multiply:
        new_value = sheet.datasets[element].loc[index, col] * increase_rate
    else:
        new_value = sheet.datasets[element].loc[index, col] + increase_rate

    if bound is not None:
        if new_value <= bound:
            sheet.datasets[element].loc[index, col] = new_value
    else:
        sheet.datasets[element].loc[index, col] = new_value


def decrease(sheet, element, index, decrease_rate, col, divide=True, bound=None):
    """Decrease the value in the dataset at position index/col.

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    element : str: 'cell' or 'face' or 'edge' or 'vert'
    index : index in the datasets[element]
    decrease_rate : rate use to divide value in the column col.
    col : column from element which apply decrease_rate.
    divide : bool: if true the current col value is divide by decrease_rate. If false it is substracted. Default divide.
    bound: lower limit of the modify value. Default None.
    """
    if divide:
        new_value = sheet.datasets[element].loc[index, col] / decrease_rate
    else:
        new_value = sheet.datasets[element].loc[index, col] - decrease_rate

    if bound is not None:
        if new_value >= bound:
            sheet.datasets[element].loc[index, col] = new_value
    else:
        sheet.datasets[element].loc[index, col] = new_value


def exchange(sheet, face, geom, remove_tri_faces=True):
    """
    Execute a type1 transition on the shorter edge of a face.

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    face : index of the face
    geom : a Geometry class
    remove_tri_faces : remove automaticaly triangular faces if existed. Default True.
    """
    edges = sheet.edge_df[sheet.edge_df["face"] == face]
    shorter = edges.length.idxmin()
    # type1_transition(sheet, shorter, 2 * min(edges.length), remove_tri_faces)
    type1_transition(sheet, shorter, epsilon=0.1, remove_tri_faces=remove_tri_faces)
    geom.update_all(sheet)


def remove(sheet, face, geom):
    """
    Removes the face and updates the geometry

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    face : index of the face
    geom : a Geometry class

    """
    remove_face(sheet, face)
    geom.update_all(sheet)


def ab_pull(sheet, face, radial_tension, distributed=False):
    """ Adds radial_tension to the face's vertices radial_tension

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    face : index of face
    radial_tension :
    distributed : bool: If true devide radial_tension by number of vertices, and apply this new radial tension to each vertices. Default not distributed.

    """
    verts = sheet.edge_df[sheet.edge_df["face"] == face]["srce"].unique()
    if distributed:
        radial_tension = radial_tension / len(verts)

    sheet.vert_df.loc[verts, "radial_tension"] += radial_tension


def increase_linear_tension(
    sheet,
    face,
    line_tension_increase,
    multiply=True,
    isotropic=True,
    angle=np.pi / 4,
    limit=100,
):
    """
    Increase edges line tension from face isotropic or according to an angle.

    Parameters
    ----------
    face : index of face
    line_tension_increase : factor for increase line tension value
    multiply : line_tension_increase is multiply or add to the current
                line_tension value. Default True.
    isotropic : all edges are increase, or only a subset of edges. Default True.
    angle : angle below edges are increase by line_tension_increase if
                isotropic is False. Default pi/4
    limit : line_tension stay below this limit value

    """
    edges = sheet.edge_df[sheet.edge_df["face"] == face]

    if isotropic:
        for index, edge in edges.iterrows():
            increase(
                sheet,
                "edge",
                edge.name,
                line_tension_increase,
                "line_tension",
                multiply,
                limit,
            )

    else:
        for index, edge in edges.iterrows():
            angle_ = np.arctan2(
                sheet.edge_df.loc[edge.name, "dx"], sheet.edge_df.loc[edge.name, "dy"]
            )

            if np.abs(angle_) < np.pi / 4:
                increase(
                    sheet,
                    "edge",
                    edge.name,
                    line_tension_increase,
                    "line_tension",
                    multiply,
                    limit,
                )


def grow(sheet, face, growth_rate, growth_col="prefered_vol"):
    """Multiplies the grow columns of face by a factor.

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    face : index of face
    growth_rate : rate use to multiply value of growth_col of face.
    growth_col : column from face dataframe which apply growth_rate.
                growth_col need to exist in face_df. Default 'prefered_vol'


    :Example:

    >>> print(sheet.face_df[face, 'prefered_vol'])
    10
    >>> grow(sheet, face, 1.7, 'prefered_vol')
    >>> print(sheet.face_df[face, 'prefered_vol'])
    17.0

    """
    warnings.warn("deprecated, use increase function")
    increase(sheet, "face", face, growth_rate, growth_col, True)


def shrink(sheet, face, shrink_rate, shrink_col="prefered_vol"):
    """Devides the shrink_col of face by a factor.

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    face : index of face
    shrink_rate : rate use to multiply value of shrink_col of face.
    shrink_col : column from face dataframe which apply shrink_rate.
                shrink_col need to exist in face_df. Default 'prefered_vol'
    """
    warnings.warn("deprecated, use decrease function")
    decrease(sheet, "face", face, shrink_rate, shrink_col, True)


def contract(
    sheet, face, contractile_increase, multiply=False, contract_col="contractility"
):
    """
    Contract the face by increasing the 'contractility' parameter
    by contractile_increase

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    face : index of face
    contractile_increase : rate use to multiply/add value of contraction_col of face.
    multiply : contractile_increase is multiply/add to the current contract_col value.
                Default False.
    contract_col : column from face dataframe which apply contractile_increase.
                contract_col need to exist in face_df. Default 'contractility'

    """
    warnings.warn("deprecated, use increase function")
    increase(sheet, "face", face, contractile_increase, contract_col, multiply)


def relax(sheet, face, relax_decrease, relax_col="contractility"):
    """
    Relax the face by decreasing the relax_col parameter
    by relax_decrease

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    face : index of face
    relax_decrease : rate use to divide value of relax_col of face.
    relax_col : column from face dataframe which apply relax_decrease.
                relax_col need to exist in face_df. Default 'contractility'

    """

    warnings.warn("deprecated, use decrease function")
    initial_contractility = 1.12
    decrease(
        sheet,
        "face",
        face,
        relax_decrease,
        col=relax_col,
        divide=True,
        bound=(initial_contractility / 2),
    )
    increase(sheet, "face", face, relax_decrease, "prefered_area", True)



"""
Input/Output module.

Supported formats:

hdf5, csv
"""
import os

from pathlib import Path


def get_sim_dir():
    """Returns the path to the root simulation directory
    as a :class:`pathlib.Path` object

    The simulation directory is stored in the  'TYSSUE_SIM_DIR'
    environement variable.


    """
    try:
        sim_dir = Path(os.environ["TYSSUE_SIM_DIR"]).absolute()
    except KeyError:
        print(
            """
No default found for the simulations directory,
you can define it by setting the "TYSSUE_SIM_DIR"
environment variable on your system (e.g in .bashrc).

We are now using the current directory as default.
            """
        )
        sim_dir = Path(os.getcwd())

    return sim_dir



import numpy as np
from functools import wraps
from itertools import count

from ..core.sheet import Sheet

from .base_topology import *
from .sheet_topology import type1_transition, remove_face
from .bulk_topology import (
    HI_transition,
    IH_transition,
    find_HIs,
    find_IHs,
    find_rearangements,
)


MAX_ITER = 10


class TopologyChangeError(ValueError):
    """ Raised when trying to assign values without
    the correct length to an epithelium dataset
    """

    pass


def all_rearangements(eptm, with_t3=True):
    """Performs rearangements (T3/HI first) until
    there are none left or MAX_ITER is reached.
    """
    for i in count():
        if i == MAX_ITER:
            return 2
        retcode = single_rearangement(eptm, with_t3=with_t3)
        if retcode:  # No transition found
            return retcode


def single_rearangement(eptm, with_t3=True):
    """Performs a single rearangement (if any) on epithelium `eptm`.

    If `with_t3` is True and there are removeable faces, will perform
    a type 3 or HI transition on one of those faces. If no such
    transition should occur, will perform a type 1 or IH transition on one
    of the edges.
    """
    edges, faces = find_rearangements(eptm)

    if len(edges):
        for i in count():
            if i == MAX_ITER:
                return 3
            if isinstance(eptm, Sheet):
                retcode = type1_transition(eptm, np.random.choice(edges))
            else:
                retcode = IH_transition(eptm, np.random.choice(edges))
            if not retcode:
                return 0

    elif len(faces) and with_t3:
        for i in count():
            if i == MAX_ITER:
                return 3
            if isinstance(eptm, Sheet):
                retcode = remove_face(eptm, np.random.choice(faces))
            else:
                retcode = HI_transition(eptm, np.random.choice(faces))
            if not retcode:
                return 0
    return 1


def auto_t1(fun):
    @wraps(fun)
    def with_rearange(*args, **kwargs):
        eptm, geom = args[:2]
        logger.debug("checking for t1s")
        res = fun(*args, **kwargs)
        shorts = find_IHs(eptm)
        if not len(shorts):
            return res
        i = 0
        while len(shorts):
            # shorter_edge = shorts[0]
            shorter_edge = np.random.choice(shorts)
            logger.info("transition on  edge %i", shorter_edge)
            if isinstance(eptm, Sheet):
                type1_transition(eptm, shorter_edge)
            else:
                IH_transition(eptm, shorter_edge)
            eptm.reset_index()
            eptm.reset_topo()
            geom.update_all(eptm)
            shorts = find_IHs(eptm)
            if len(shorts) and shorts[0] == shorter_edge:
                # IH transition did not work, skipping
                logger.info("removing shorter edge %i from edge list", shorter_edge)
                shorts = shorts[1:]
            i += 1
            if i > MAX_ITER:
                break
        if eptm.position_buffer is not None:
            logger.info("out T1 changed buffer")
            eptm.position_buffer = eptm.vert_df[eptm.coords].copy()
        logger.info("performed %i T1", i)
        return res

    return with_rearange


def auto_t3(fun):
    @wraps(fun)
    def with_rearange(*args, **kwargs):
        eptm, geom = args[:2]
        logger.debug("checking for t3s")
        res = fun(*args, **kwargs)
        tri_faces = find_HIs(eptm)
        if not len(tri_faces):
            return res
        i = 0
        while len(tri_faces):
            smaller_face = tri_faces[0]
            logger.debug("Performing t3 on face %d", smaller_face)
            if isinstance(eptm, Sheet):
                remove_face(eptm, smaller_face)
            else:
                HI_transition(eptm, smaller_face)
            eptm.reset_index()
            eptm.reset_topo()
            geom.update_all(eptm)
            tri_faces = find_HIs(eptm)
            i += 1
            if i > MAX_ITER:
                break
        if eptm.position_buffer is not None:
            logger.info("out T3 changed buffer")
            eptm.position_buffer = eptm.vert_df[eptm.coords].copy()
        logger.info("performed %i T3", i)
        return res

    return with_rearange



import logging
import numpy as np

from ..geometry.bulk_geometry import MonolayerGeometry
from ..core.sheet import Sheet
from ..geometry.sheet_geometry import SheetGeometry
from .bulk_topology import get_division_vertices
from .bulk_topology import cell_division as bulk_division
from .sheet_topology import type1_transition as sheet_t1
from .sheet_topology import get_division_edges as sheet_division_edges


logger = logging.getLogger(name=__name__)


def cell_division(monolayer, mother, orientation="vertical", psi=0):
    """
    Divides the cell mother in the monolayer.

    Parameters
    ----------
    * monolayer: a :class:`Monolayer` instance
    * mother: int, the index of the cell to devide
    * orientation: str, {"vertical" | "horizontal"}
      if "horizontal", performs a division in the equatorial
      plane of the cell. If "vertical" (the default), performs
      a division along the basal-apical axis of the cell
    * psi: float, default 0
      extra rotation angle of the division plane
      around the basal-apical plane

    Returns
    -------
    * daughter: int, the index of the daughter cell
    """

    ab_axis = MonolayerGeometry.basal_apical_axis(monolayer, mother)
    plane_normal = np.asarray(ab_axis)

    if orientation == "horizontal":
        vertices = get_division_vertices(
            monolayer, mother=mother, plane_normal=plane_normal
        )
    elif orientation == "vertical":
        apical_sheet = monolayer.get_sub_sheet("apical")
        m_apical_face = monolayer.edge_df[
            (monolayer.edge_df["cell"] == mother)
            & (monolayer.edge_df["segment"] == "apical")
        ]["face"].iloc[0]
        apical_edges = sheet_division_edges(apical_sheet, m_apical_face, SheetGeometry)
        basal_edges = []
        for ae in apical_edges[::-1]:
            basal_edges.append(find_basal_edge(monolayer, ae))
        division_edges = list(apical_edges) + basal_edges
        vertices = get_division_vertices(monolayer, division_edges=division_edges)
    else:
        raise ValueError(
            """orientation argument not understood,
should be either "horizontal" or "vertical", not {}""".format(
                orientation
            )
        )

    daughter = bulk_division(monolayer, mother, MonolayerGeometry, vertices)

    # Correct segment assignations for the septum
    septum = monolayer.face_df.index[-2:]
    septum_edges = monolayer.edge_df.index[-2 * len(vertices) :]
    if orientation == "vertical":
        monolayer.face_df.loc[septum, "segment"] = "lateral"
        monolayer.edge_df.loc[septum_edges, "segment"] = "lateral"
        _assign_vert_segment(monolayer, vertices)

    elif orientation == "horizontal":
        monolayer.face_df.loc[septum[0], "segment"] = "apical"
        monolayer.face_df.loc[septum[1], "segment"] = "basal"
        monolayer.edge_df.loc[septum_edges[: len(vertices)], "segment"] = "apical"
        monolayer.edge_df.loc[septum_edges[len(vertices) :], "segment"] = "basal"
        monolayer.vert_df.loc[vertices, "segment"] = "apical"

    return daughter


def _assign_vert_segment(monolayer, vertices):

    for v in vertices:
        segs = set(monolayer.edge_df[monolayer.edge_df["srce"] == v]["segment"])
        if "apical" in segs:
            monolayer.vert_df.loc[v, "segment"] = "apical"
        elif "basal" in segs:
            monolayer.vert_df.loc[v, "segment"] = "basal"
        else:
            monolayer.vert_df.loc[v, "segment"] = "lateral"


def find_basal_edge(monolayer, apical_edge):
    """Returns the basal edge parallel to the apical edge passed
    in argument.

    Parameters
    ----------
    monolayer: a :class:`Monolayer` instance

    """
    srce, trgt, cell = monolayer.edge_df.loc[apical_edge, ["srce", "trgt", "cell"]]
    cell_edges = monolayer.edge_df[monolayer.edge_df["cell"] == cell]
    srce_segment = monolayer.vert_df.loc[cell_edges["srce"].values, "segment"]
    srce_segment.index = cell_edges.index
    trgt_segment = monolayer.vert_df.loc[cell_edges["trgt"].values, "segment"]
    trgt_segment.index = cell_edges.index
    b_trgt, = cell_edges[
        (srce_segment == "apical")
        & (trgt_segment == "basal")
        & (cell_edges["srce"] == srce)
    ]["trgt"]
    b_srce, = cell_edges[
        (srce_segment == "basal")
        & (trgt_segment == "apical")
        & (cell_edges["trgt"] == trgt)
    ]["srce"]
    b_edge, = cell_edges[
        (cell_edges["srce"] == b_srce) & (cell_edges["trgt"] == b_trgt)
    ].index
    return b_edge


def type1_transition(monolayer, apical_edge, epsilon=0.1):
    """Performs a type 1 transition on the apical and basal meshes
    """
    v0_a, v1_a, fb_a, cb_a = monolayer.edge_df.loc[
        apical_edge, ["srce", "trgt", "face", "cell"]
    ]
    basal_edge = find_basal_edge(monolayer, apical_edge)
    v0_b, v1_b, fb_b, cb_b = monolayer.edge_df.loc[
        basal_edge, ["srce", "trgt", "face", "cell"]
    ]
    if monolayer.face_df.loc[fb_a, "num_sides"] < 4:
        logger.warning(
            """Face %s has 3 sides,
type 1 transition is not allowed""",
            fb_a,
        )
        return



"""
Base gradients for bulk geometry
"""
import numpy as np
import pandas as pd


def volume_grad(eptm):

    coords = eptm.coords
    cell_pos = eptm.edge_df[["c" + c for c in coords]].values
    face_pos = eptm.edge_df[["f" + c for c in coords]].values
    srce_pos = eptm.edge_df[["s" + c for c in coords]].values
    trgt_pos = eptm.edge_df[["t" + c for c in coords]].values

    grad_v_srce = np.cross((trgt_pos - cell_pos), (face_pos - cell_pos)) / 4
    grad_v_trgt = -np.cross((srce_pos - cell_pos), (face_pos - cell_pos)) / 4
    return (
        pd.DataFrame(grad_v_srce, index=eptm.edge_df.index, columns=eptm.coords),
        pd.DataFrame(grad_v_trgt, index=eptm.edge_df.index, columns=eptm.coords),
    )


def lumen_volume_grad(eptm):
    """
    Calculates the gradient for the volume enclosed by the epithelium.

    For a monolayer, it will by default compute the volume enclosed
    by the basal side (edges whose 'segment' column is "basal").
    If the polarity is reversed and the apical side faces the lumen,
    this can be changed by setting eptm.settings["lumen_side"] to 'apical'

    """

    coords = eptm.coords
    if "segment" in eptm.edge_df:
        lumen_side = eptm.settings.get("lumen_side", "basal")
        basal_edges = eptm.edge_df[eptm.edge_df.segment == lumen_side]
        face_pos = basal_edges[["f" + c for c in coords]].values
        srce_pos = basal_edges[["s" + c for c in coords]].values
        trgt_pos = basal_edges[["t" + c for c in coords]].values

        grad_v_srce = pd.DataFrame(
            np.zeros((eptm.Ne, 3)), index=eptm.edge_df.index, columns=eptm.coords
        )
        grad_v_trgt = pd.DataFrame(
            np.zeros((eptm.Ne, 3)), index=eptm.edge_df.index, columns=eptm.coords
        )

        grad_v_srce.loc[basal_edges.index] = -np.cross((trgt_pos), (face_pos)) / 4
        grad_v_trgt.loc[basal_edges.index] = np.cross((srce_pos), (face_pos)) / 4

        return grad_v_srce, grad_v_trgt
    else:
        face_pos = eptm.edge_df[["f" + c for c in coords]].values
        srce_pos = eptm.edge_df[["s" + c for c in coords]].values
        trgt_pos = eptm.edge_df[["t" + c for c in coords]].values

        grad_v_srce = np.cross((trgt_pos), (face_pos)) / 4
        grad_v_trgt = -np.cross((srce_pos), (face_pos)) / 4
        return (
            pd.DataFrame(grad_v_srce, index=eptm.edge_df.index, columns=eptm.coords),
            pd.DataFrame(grad_v_trgt, index=eptm.edge_df.index, columns=eptm.coords),
        )
